/*
 * @(#) SubCoordinator.java 
 *
 * JOTM: Java Open Transaction Manager 
 *
 * This project is developed inside the ObjectWeb Consortium,
 * http://www.objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Contributor(s):
 * 01/11/06 Christophe Ney cney@batisseurs.com for Lutris Technologies
 *          Added ResourceManagerListener mechanism to remove ThreadData
 *          dependency.
 *
 * 02/06/18 Marek Prochazka
 *          In addSynchronization() synchronizations are processed the
 *          same way as resources. We add synchro although transaction is
 *          marked rollback-only to be able to send beforeCompletion
 *
 * --------------------------------------------------------------------------
 * $Id: SubCoordinator.java,v 1.15 2003/05/14 11:06:26 jmesnil Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jotm;

import java.rmi.RemoteException;

import java.util.List;
import java.util.Vector;
import javax.transaction.*;
import javax.transaction.xa.XAException;
import javax.transaction.xa.XAResource;
import javax.transaction.xa.Xid;
import javax.rmi.PortableRemoteObject;

/**
 * Log associated to this transaction coordinator
 */
class SLog {

    private Vector loggedResources = new Vector();
    int decision;
    static final int DECISION_TO_COMMIT = 1;
    static final int DECISION_TO_ROLLBACK = 2;

    public void addResource(XAResource res) {
        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("res="+ res);
        }
        loggedResources.addElement(res);
    }

    /**
     * @return a List of logged Resources
     */
    public List getLoggedResources() {
        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("logged resources="+ loggedResources);
        }
        return loggedResources;
    }

    public void flushLog(int decide) {
        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("decide="+ decide);
        }
        decision = decide;

        // XXX serialize log on disk
    }

    public void forgetLog() {
        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("forget log");
        }
        // XXX remove file on disk
    }
}

/**
 * This object is the local coordinator. It may be registered as
 * sub-coordinator in case of distributed transaction, so it must
 * be callable remotely and implement Resource
 */
public class SubCoordinator extends PortableRemoteObject implements Resource {

    // ------------------------------------------------------------------
    // Object state
    // ------------------------------------------------------------------

    /**
     * @serial
     */
    private TransactionImpl tx = null;

    /**
     * List of Synchronization objects
     * @serial
     */
    private Vector synchroList = new Vector();

    /**
     * List of XAResource objects
     * @serial
     */
    private Vector resourceList = new Vector();

    /**
     * Keep a reference on TransactionManager
     * @serial
     */
    private TransactionManager tm;

    /**
     * @serial
     */
    private Xid xid = null;

    /**
     * @serial
     */
    private SLog log = null;

    /**
     * javax.transaction.Status     
     * <dl>
     *    <dt>ACTIVE</dt><dd>transaction started, commit phase not started</dd>
     *    <dt>PREPARING</dt><dd>prepare is being sent to resources</dd>
     *    <dt>PREPARED</dt><dd>prepare is done. Must commit now</dd>
     *    <dt>COMMITTING</dt><dd>commit is being sent to resources</dd>
     *    <dt>COMMITTED</dt><dd>commit was successful</dd>
     *    <dt>ROLLING_BACK</dt><dd>not used</dd>
     *    <dt>MARKED_ROLLBACK</dt><dd>setRollbackOnly has been called</dd>
     *    <dt>ROLLEDBACK</dt>
     *    <dd>transaction has been rolled back (or prepare returned "vote_rollback")</dd>
     *    <dt>UNKNOWN</dt><dd>commit raised heuristics</dd>
     *     <dt>NO_TRANSACTION</dt><dd>cannot get the status value from JTM</dd>
     * </dl>
     * @serial
     */
    private int status = Status.STATUS_ACTIVE;

    /**
     * @serial
     */
    private boolean beforeCompletionDone = false;

    // ------------------------------------------------------------------
    // Constructors
    // ------------------------------------------------------------------

    /**
     * constructor used by TransactionImpl
     */
    SubCoordinator(TransactionImpl tx, Xid xid) throws RemoteException {
        TraceTm.jta.debug("tx="+ tx +",  xid="+ xid);
        this.tx = tx;
        this.xid = xid;
        this.tm = Current.getTransactionManager();

        // increment counter for management
        Current.getCurrent().incrementBeginCounter();
    }

    // ------------------------------------------------------------------
    // Resource interface
    // ------------------------------------------------------------------

    /** 
     * phase 1 of the 2PC.
     *
     * @return int vote commit, rollback, or readonly.
     */
    public int prepare() throws RemoteException {
        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("status="+ StatusHelper.getStatusName(status));
        }
        

		try {
			tx.doDetach(XAResource.TMSUCCESS);
		} catch (SystemException e) {
			if (TraceTm.jta.isDebugEnabled()) {
				TraceTm.jta.debug("error when detaching xaresources", e);
			}
		}

        switch (status) {
            case Status.STATUS_MARKED_ROLLBACK :
                doBeforeCompletion(false);
                doRollback();
                return Resource.VOTE_ROLLBACK;
            case Status.STATUS_COMMITTED :
                return Resource.VOTE_COMMIT;
            default :
                doBeforeCompletion(true);
                break;
        }

        // Recheck Status after doBeforeCompletion
        if (status == Status.STATUS_MARKED_ROLLBACK) {
            TraceTm.jotm.info(
                "Rollback during beforeCompletion in SubCoordinator.prepare");
            doRollback();
            return Resource.VOTE_ROLLBACK;
        }

        int ret = doPrepare();
        if (ret == Resource.VOTE_READONLY) {
            // Transaction completed for this Resource
            doAfterCompletion();
        }
        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("vote = " + ret);
        }
        return ret;
    }

    /** 
     * rollback transaction
     */
    public void rollback() throws RemoteException {
		if (TraceTm.jta.isDebugEnabled()) {
			TraceTm.jta.debug("status="+ StatusHelper.getStatusName(status));
		}
       
		try {
			tx.doDetach(XAResource.TMSUCCESS);
		} catch (SystemException e) {
			if (TraceTm.jta.isDebugEnabled()) {
				TraceTm.jta.debug("error when detaching xaresources", e);
			}
		}

        switch (status) {
            case Status.STATUS_ACTIVE :
            case Status.STATUS_MARKED_ROLLBACK :
            case Status.STATUS_ROLLING_BACK:
            case Status.STATUS_PREPARED:
                break;
            case Status.STATUS_ROLLEDBACK :
                return;
            default :
                TraceTm.jotm.error(
                    "rollback: bad status: "
                        + StatusHelper.getStatusName(status));
                return;
        }

        doBeforeCompletion(false);
        doRollback();
    }

    /** 
     * phase 2 of the 2PC.
     */
    public void commit() throws RemoteException {
        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("status="+ StatusHelper.getStatusName(status));
        }

        switch (status) {
            case Status.STATUS_PREPARED :
                break;
            default :
                TraceTm.jotm.error(
                    "commit: bad status: "
                        + StatusHelper.getStatusName(status));
                return;
        }

        doCommit();
    }

    /** 
     * commit 1 phase. Called either from JTM (distributed transaction) or
     * from Transaction.commit (local transaction).
     */
    public void commit_one_phase() throws RemoteException {
		if (TraceTm.jta.isDebugEnabled()) {
			TraceTm.jta.debug("status="+ StatusHelper.getStatusName(status));
		}

		try {
            tx.doDetach(XAResource.TMSUCCESS);
        } catch (SystemException e) {
            if (TraceTm.jta.isDebugEnabled()) {
                TraceTm.jta.debug("error when detaching xaresources", e);
            }
        }
        
        switch (status) {
            case Status.STATUS_ROLLEDBACK :
                throw new TransactionRolledbackException();
            case Status.STATUS_MARKED_ROLLBACK :
                doBeforeCompletion(false);
                doRollback();
                throw new TransactionRolledbackException();
            case Status.STATUS_COMMITTED :
                return;
            default :
                doBeforeCompletion(true);
                break;
        }

        // Recheck Status after doBeforeCompletion
        if (status == Status.STATUS_MARKED_ROLLBACK) {
            TraceTm.jotm.info(
                "Rollback during beforeCompletion in SubCoordinator.commit_one_phase");
            doRollback();
            throw new TransactionRolledbackException();
        }

        // only 1 Resource => 1 phase commit
        if (resourceList.size() == 1) {
            doOnePhaseCommit();
            return;
        }

        // 2 phase commit
        int vote = doPrepare();

        switch (vote) {
            case Resource.VOTE_COMMIT :
                doCommit();
                break;
            case Resource.VOTE_READONLY :
                doAfterCompletion();
                break;
            case Resource.VOTE_ROLLBACK :
                doRollback();
                throw new TransactionRolledbackException();
        }
    }

    /** 
     * forget heuristics about this transaction.
     */
    public void forget() throws RemoteException {
        TraceTm.jta.debug("SubCoordinator.forget()");
        doForget();
    }

    // ------------------------------------------------------------------
    // Other public methods (interface exposed to Transaction object)
    // ------------------------------------------------------------------

    /**
     * add a Synchronization to the list
     *
     * @param synchro The javax.transaction.Synchronization object for the 
     *    transaction associated with the target object
     *
     * @exception RollbackException Thrown to indicate that
     *    the transaction has been marked for rollback only.
     *
     * @exception IllegalStateException Thrown if the transaction in the 
     *    target object is in prepared state or the transaction is inactive.
     *
     */
    public void addSynchronization(Synchronization synchro)
        throws RollbackException, IllegalStateException {
        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("synchro="+ synchro);
            TraceTm.jta.debug("status="+ StatusHelper.getStatusName(status));
        }

        //check status: should be ACTIVE.
        boolean markedRollback = false;
        switch (status) {
            case Status.STATUS_MARKED_ROLLBACK :
            case Status.STATUS_ROLLEDBACK :
                markedRollback = true;
                break;
            case Status.STATUS_ACTIVE :
                break;
            default :
                String errorMsg =
                    "addSynchronization: bad status = "
                        + StatusHelper.getStatusName(status);
                TraceTm.jotm.error(errorMsg);
                throw new IllegalStateException(errorMsg);
        }

        // Add synchro to the list of local synchros
        synchroList.addElement(synchro);

        // If transaction marked rollback only, we add synchro but we throw
        // the correct exception because nobody can presume what will be done in the
        // Synchronization object, so we must send it beforeCompletion and afterCompletion
        // even in case of rollback.
        if (markedRollback) {
            TraceTm.jta.debug(
                "SubCoordinator.addSynchronization: transaction rollback only");
            throw new RollbackException();
        }

    }

    /**
     * add a XAResource to the list
     *
     * @param the XAResource to register
     *
     * @return true if this datasource was already known
     *
     * @exception RollbackException Thrown to indicate that
     *    the transaction has been marked for rollback only.
     *
     * @exception IllegalStateException Thrown if the transaction in the 
     *    target object is in prepared state or the transaction is inactive.
     */
    public boolean addResource(XAResource xares)
        throws RollbackException, IllegalStateException {
        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("xares="+ xares);
            TraceTm.jta.debug("status="+ StatusHelper.getStatusName(status));
        }

        //check status: should be ACTIVE.
        boolean markedRollback = false;
        switch (status) {
            case Status.STATUS_MARKED_ROLLBACK :
                markedRollback = true;
                break;
            case Status.STATUS_ACTIVE :
                break;
            default :
                String errorMsg =
                    "SubCoordinator.addResource: bad status="
                        + StatusHelper.getStatusName(status);
                TraceTm.jotm.error(errorMsg);
                throw new IllegalStateException(errorMsg);
        }

        // Check if this DataSource is already known
        // -> we must register only ONE resource per DataSource.
        boolean found = false;
        for (int i = 0; i < resourceList.size(); i++) {
            XAResource res = (XAResource) resourceList.elementAt(i);
            try {
                if (res.isSameRM(xares)) {
                    found = true;
                    break;
                }
            } catch (XAException e) {
                TraceTm.jotm.error("Exception on isSameRM:", e);
            }
        }

        // add this XAResource to the list
        if (found == false) {
            TraceTm.jta.debug("new XAResource added to the list");
            resourceList.addElement(xares);
        }

        // If transaction marked rollback only, we enlist resource but we throw
        // the correct exception. It is important to enlist the Resource because
        // if we don't, the database would be updated although transaction has been
        // rolled back.
        if (markedRollback) {
            TraceTm.jta.debug(
                "SubCoordinator.addResource: transaction rollback only");
            throw new RollbackException();
        }

        return found;
    }

    /**
     * return the status of this transaction
     */
    public int getStatus() {
        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("status="+ StatusHelper.getStatusName(status));
        }
        return status;
    }

    /**
     * set the transaction "rollback only"
     */
    public void setRollbackOnly() {        
        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("status="+ StatusHelper.getStatusName(status));
        }
        switch (status) {
            case Status.STATUS_ACTIVE :
            case Status.STATUS_UNKNOWN :
            case Status.STATUS_PREPARING :
                status = Status.STATUS_MARKED_ROLLBACK;
                break;
            case Status.STATUS_MARKED_ROLLBACK :
            case Status.STATUS_ROLLING_BACK :
                break;
            case Status.STATUS_PREPARED :
            case Status.STATUS_COMMITTED :
            case Status.STATUS_ROLLEDBACK :
            case Status.STATUS_NO_TRANSACTION :
            case Status.STATUS_COMMITTING :
                TraceTm.jotm.error("Cannot set transaction as rollback only");
                TraceTm.jotm.error(
                    "Bad status="+ StatusHelper.getStatusName(status));
                break;
        }
    }

    // ------------------------------------------------------------------
    // private methods
    // ------------------------------------------------------------------

    /**
     * forget Transaction
     */
    private void doForget() throws RemoteException {
        TraceTm.jta.debug("SubCoordinator.doForget()");

        boolean exception = false;

        for (int i = 0; i < resourceList.size(); i++) {
            // XXX Tell this Resource to forget the transaction
            // Remove from the list ?
            XAResource xar = (XAResource) resourceList.elementAt(i);
            try {
                xar.forget(xid);
            } catch (XAException e) {
                exception = true;
            }
        }
        if (exception) {
            throw new RemoteException("XAException on forget");
        }
        unexportObject(this);
    }

    /**
     * 2PC phase 1 (prepare)
     * Basically, send prepare on each XAResource and
     * collect the results.
     */
    private int doPrepare() throws RemoteException {
        TraceTm.jta.debug("SubCoordinator.doPrepare()");

        int ret = VOTE_READONLY;
        int errors = 0;

        // No resource
        if (resourceList.size() == 0) {
            // increment counter for management
            Current.getCurrent().incrementCommitCounter();
            status = Status.STATUS_COMMITTED;
            return ret;
        }

        // Creates a log for that transaction, where we will add all the
        // resources that replied VOTE_COMMIT to prepare.
        // Do not flush the log on disk before decision to commit.
        log = new SLog();

        // Sends prepare to each resource while no error
        // In case of prepare on sub-coord. we may have only 1 resource.
        status = Status.STATUS_PREPARING;
        for (int i = 0; i < resourceList.size(); i++) {
            XAResource res = (XAResource) resourceList.elementAt(i);
            if (errors > 0) {
                try {
                    res.rollback(xid);
                } catch (XAException e) {
                    // Nothing to do ?
                }
            } else {
                try {
                    switch (res.prepare(xid)) {
                        case XAResource.XA_OK :
                            log.addResource(res);
                            ret = VOTE_COMMIT;
                            break;
                        case XAResource.XA_RDONLY :
                            break;
                    }
                } catch (XAException e) {
                    ret = VOTE_ROLLBACK;
                    errors++;
                }
            }
        }

        // Update the status, depending on vote result
        // If all resources returned READ_ONLY, we can forget the transaction
        switch (ret) {
            case VOTE_READONLY :
                // increment counter for management
                Current.getCurrent().incrementCommitCounter();
                status = Status.STATUS_COMMITTED;
                break;
            case VOTE_COMMIT :
                status = Status.STATUS_PREPARED;
                break;
            case VOTE_ROLLBACK :
                status = Status.STATUS_ROLLING_BACK;
                break;
        }

        // return the global vote
        return ret;
    }

    /**
     * 2PC - phase 2 (commit)
     * XXX See JTM for heuristics management
     */
    private int doCommit() throws RemoteException {

        TraceTm.jta.debug("SubCoordinator.doCommit()");

        // First check that a log is initialized
        if (log == null) {
            TraceTm.jotm.error("doCommit: no log");
            return -1;
        }

        // Status Transaction = committing
        status = Status.STATUS_COMMITTING;

        // Send commit to each resource prepared
        int errors = 0;
        int commitnb = 0;
        int heuristicnb = 0;
        List loggedResources = log.getLoggedResources();
        for (int i = 0; i < loggedResources.size(); i++) {
            XAResource res = (XAResource) loggedResources.get(i);

            // aborts transaction if error and no resource committed yet
            if (commitnb == 0 && errors > 0) {
                try {
                    res.rollback(xid);
                } catch (XAException e) {
                    commitnb++;
                    heuristicnb++;
                }
                continue;
            }

            // commit resource
            try {
                res.commit(xid, false);
            } catch (XAException e) {
                errors++;
                if (commitnb > 0)
                    heuristicnb++;
            }
        }

        if (errors == 0) {
            // increment counter for management
            Current.getCurrent().incrementCommitCounter();
            // Everything's fine.
            status = Status.STATUS_COMMITTED;
            doAfterCompletion();
            log.forgetLog();
            return 0;
        }

        if (heuristicnb == 0) {
            // Transaction has been eventually rolled back
            status = Status.STATUS_ROLLEDBACK;
            doAfterCompletion();
            log.forgetLog();
            throw new TransactionRolledbackException();
        }

        // XXX Log heuristics if errors
        status = Status.STATUS_UNKNOWN;
        doAfterCompletion();
        return -1;
    }

    /**
     * 1PC (commit one phase)
     */
    private void doOnePhaseCommit() throws RemoteException {

        TraceTm.jta.debug("SubCoordinator.doOnePhaseCommit()");

        // Only 1 resource: commit with onePhase=true
        status = Status.STATUS_COMMITTING;
        XAResource res = (XAResource) resourceList.elementAt(0);
        try {

            // increment counter for management
            Current.getCurrent().incrementCommitCounter();
            res.commit(xid, true);
            status = Status.STATUS_COMMITTED;
        } catch (XAException e) {
            status = Status.STATUS_UNKNOWN;
            if (e.errorCode == XAException.XA_RBROLLBACK) {
                throw new TransactionRolledbackException("XAException:"+ e);
            }
            throw new RemoteException("XAException:"+ e);
        } finally {
            doAfterCompletion();
        }
    }

    /**
     * Rollback every resource involved
     */
    private void doRollback() throws RemoteException {

        TraceTm.jta.debug("SubCoordinator.doRollback()");

        status = Status.STATUS_ROLLEDBACK;
        boolean hcomm = false;

        // rollback each resource
        for (int i = 0; i < resourceList.size(); i++) {
            XAResource res = (XAResource) resourceList.elementAt(i);
            try {
                res.rollback(xid);
            } catch (XAException e) {
                switch (e.errorCode) {
                    case XAException.XA_HEURCOM :
                        hcomm = true;
                    default :
                        TraceTm.jotm.error(
                            "Got XAException from resource: "+ e);
                        throw new RemoteException(
                            "rollback: Unexpected XAException:"+ e.errorCode);
                }
            }
        }

        doAfterCompletion();
        // increment counter for management
        Current.getCurrent().incrementRollbackCounter();

        // raise exception if error on a resource
        if (hcomm) {
            throw new HeuristicCommit();
        }
    }

    /**
     * before completion
     *
     * @param boolean true if completion ok, false if rollback
     */
    private void doBeforeCompletion(boolean committing) {
        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("committing="+ committing);
        }

        if (beforeCompletionDone)
            return;

        // Unset the timer for this transaction, if any
        tx.unsetTimer();

        // For each synchro, send beforeCompletion (not if rollback)
        if (committing && synchroList.size() > 0) {
            // We must be in the correct transaction context
            // See JTA spec. page 13 (3.3.2)
            // because at beforeCompletion, the bean will write its cache

            // Check the trivial case where we already have the correct tx context
            Transaction mytx = null;
            boolean suspended = false;
            boolean resumed = false;
            try {
                mytx = tm.getTransaction();
            } catch (SystemException e) {
                if (TraceTm.jta.isDebugEnabled())
                    TraceTm.jta.debug("Cannot get transaction:"+ e);
            }

            // Suspend if another tx context
            if (mytx != null && mytx.equals(tx) == false) {
                try {
                    tm.suspend();
                    suspended = true;
                } catch (SystemException e) {
                    if (TraceTm.jta.isDebugEnabled())
                        TraceTm.jta.debug("Cannot suspend transaction:"+ e);
                }
            }

            // Resume the good tx context
            if (mytx == null || suspended) {
                try {
                    tm.resume(tx);
                    resumed = true;
                } catch (SystemException e) {
                    if (TraceTm.jta.isDebugEnabled())
                        TraceTm.jta.debug("Cannot resume transaction:"+ e);
                } catch (InvalidTransactionException e) {
                    if (TraceTm.jta.isDebugEnabled())
                        TraceTm.jta.debug("Cannot resume transaction:"+ e);
                } catch (IllegalStateException e) {
                    if (TraceTm.jta.isDebugEnabled())
                        TraceTm.jta.debug("Cannot resume transaction:"+ e);
                }
            }

            // Call the synchronizations
            // beforeCompletion may set the TX rollbackonly if something goes wrong
            for (int i = 0; i < synchroList.size(); i++) {
                Synchronization sync =
                    (Synchronization) synchroList.elementAt(i);
                sync.beforeCompletion();
            }

            // reset context as it was before
            if (resumed) {
                try {
                    tm.suspend();
                } catch (SystemException e) {
                    if (TraceTm.jta.isDebugEnabled()) {
                        TraceTm.jta.debug("Cannot suspend transaction:"+ e);
                    }
                }
            }
            if (suspended) {
                try {
                    tm.resume(mytx);
                    resumed = true;
                } catch (SystemException e) {
                    if (TraceTm.jta.isDebugEnabled()) {
                        TraceTm.jta.debug("Cannot resume transaction:"+ e);
                    }
                } catch (InvalidTransactionException e) {
                    if (TraceTm.jta.isDebugEnabled()) {
                        TraceTm.jta.debug("Cannot resume transaction:"+ e);
                    }
                } catch (IllegalStateException e) {
                    if (TraceTm.jta.isDebugEnabled()) {
                        TraceTm.jta.debug("Cannot resume transaction:"+ e);
                    }
                }
            }
        }

        beforeCompletionDone = true;
    }
    
    /**
     * after completion
     */
    private void doAfterCompletion() {
        TraceTm.jta.debug("SubCoordinator.doAfterCompletion()");

        // For each synchro, send afterCompletion
        /// CompletedTransactionListener has been replaced by Synchronization
        for (int i = 0; i < synchroList.size(); i++) {
            Synchronization sync = (Synchronization) synchroList.elementAt(i);
            sync.afterCompletion(status);
        }

        // Forget this transaction.
        // LATER: 
        // - Should not forget it in case of heuristics (for recovery)
        // - May be this could be deferred in case of retry from a client: use a timer.
        Current.getCurrent().forgetTx(tx.getXid());
        if (TraceTm.jta.isDebugEnabled()) {
            TraceTm.jta.debug("SubCoordinator unexported [subcoord=" + this + "]");
        }
        try {
            unexportObject(this);
        } catch (Exception e) {}
    }
}
